/**
 * @file
 *
 * Transmission Control Protocol, outgoing traffic
 *
 * The output functions of TCP.
 *
 */

/*
 * Copyright (c) 2001-2004 Swedish Institute of Computer Science.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
 * SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * This file is part of the lwIP TCP/IP stack.
 *
 * Author: Adam Dunkels <adam@sics.se>
 *
 */

#include "lwip/def.h"
#include "lwip/opt.h"

#include "lwip/mem.h"
#include "lwip/memp.h"
#include "lwip/sys.h"

#include "lwip/ip_addr.h"
#include "lwip/netif.h"

#include "lwip/inet.h"
#include "lwip/tcp.h"

#include "lwip/stats.h"

#if LWIP_TCP

/* Forward declarations.*/
static void tcp_output_segment(struct tcp_seg *seg, struct tcp_pcb *pcb);
u8_t tcp_output_segmentxm(struct tcp_seg *seg, struct tcp_pcb *pcb, u8_t sentmulti);
err_t tcp_outputxm(struct tcp_pcb *pcb);
void tcp_rexmitxm(struct tcp_pcb *pcb);

err_t
tcp_send_ctrl(struct tcp_pcb *pcb, u8_t flags)
{
  return tcp_enqueue(pcb, NULL, 0, flags, 1, NULL, 0);

}

/*
 * NB. tcp_write() enqueues data for sending, but does not send it
 * straight away.  It waits in the expectation of more data being sent
 * soon (as it can send them more efficiently by combining them
 * together).  To prompt the system to send data now, call
 * tcp_output() after calling tcp_write().
 */

err_t
tcp_writexm(struct tcp_pcb *pcb, const void *arg, u16_t len, u8_t copy)
{

  LWIP_DEBUGF(TCPXM_DEBUG, ("tcp_writexm: about to enqueue %u bytes\n", len));

  for(;pcb!=NULL;pcb=pcb->nextm) {
    if(len > 0) {
      LWIP_DEBUGF(TCP_OUTPUT_DEBUG, ("tcp_writexm(pcb=%p, arg=%p, len=%u, copy=%d)\n", (void *)pcb, arg, len, copy));
      tcp_enqueue(pcb, (void *)arg, len, 0, copy, NULL, 0);
    }
  }

  return ERR_OK;
}
/*-----------------------------------------------------------------------------------*/
err_t
tcp_write(struct tcp_pcb *pcb, const void *arg, u16_t len, u8_t copy)
{
  LWIP_DEBUGF(TCP_OUTPUT_DEBUG, ("tcp_write(pcb=%p, arg=%p, len=%u, copy=%d)\n", (void *)pcb,
    arg, len, (unsigned int)copy));
  if (pcb->state == SYN_SENT ||
     pcb->state == SYN_RCVD ||
     pcb->state == ESTABLISHED ||
     pcb->state == CLOSE_WAIT) {
    if (len > 0) {
      return tcp_enqueue(pcb, (void *)arg, len, 0, copy, NULL, 0);
    }
    return ERR_OK;
  } else {
    LWIP_DEBUGF(TCP_OUTPUT_DEBUG | DBG_STATE | 3, ("tcp_write() called in invalid state\n"));
    return ERR_CONN;
  }
}

err_t
tcp_enqueue(struct tcp_pcb *pcb, void *arg, u16_t len,
      u8_t flags, u8_t copy,
            u8_t *optdata, u8_t optlen)
{
  struct pbuf *p;
  struct tcp_seg *seg, *useg, *queue;
  u32_t left, seqno;
  u16_t seglen;
  void *ptr;
  u8_t queuelen;

  LWIP_DEBUGF(TCP_OUTPUT_DEBUG, ("tcp_enqueue(pcb=%p, arg=%p, len=%u, flags=%x, copy=%u)\n",
    (void *)pcb, arg, len, (unsigned int)flags, (unsigned int)copy));
  left = len;
  ptr = arg;
  /* fail on too much data */
  if (len > pcb->snd_buf) {
    LWIP_DEBUGF(TCP_OUTPUT_DEBUG | 3, ("tcp_enqueue: too much data (len=%u > snd_buf=%u)\n", len, pcb->snd_buf));
    return ERR_MEM;
  }

  /* seqno will be the sequence number of the first segment enqueued
   * by the call to this function. */
  seqno = pcb->snd_lbb;

  queue = NULL;
  LWIP_DEBUGF(TCP_QLEN_DEBUG, ("tcp_enqueue: queuelen: %u\n", (unsigned int)pcb->snd_queuelen));

  /* Check if the queue length exceeds the configured maximum queue
   * length. If so, we return an error. */
  queuelen = pcb->snd_queuelen;
  if (queuelen >= TCP_SND_QUEUELEN) {
    LWIP_DEBUGF(TCP_OUTPUT_DEBUG | 3, ("tcp_enqueue: too long queue %u (max %u)\n", queuelen, TCP_SND_QUEUELEN));
    goto memerr;
  }

  if (pcb->snd_queuelen != 0) {
    LWIP_ASSERT("tcp_enqueue: valid queue length", pcb->unacked != NULL ||
    pcb->unsent != NULL);
  }

  seg = useg = NULL;
  seglen = 0;

  /* First, break up the data into segments and tuck them together in
   * the local "queue" variable. */
  while (queue == NULL || left > 0) {

    /* The segment length should be the MSS if the data to be enqueued
     * is larger than the MSS. */
    seglen = left > pcb->mss? pcb->mss: left;

    /* Allocate memory for tcp_seg, and fill in fields. */
    seg = memp_malloc(MEMP_TCP_SEG);
    if (seg == NULL) {
      LWIP_DEBUGF(TCP_OUTPUT_DEBUG | 2, ("tcp_enqueue: could not allocate memory for tcp_seg\n"));
      goto memerr;
    }
    seg->next = NULL;
    seg->p = NULL;

    if (queue == NULL) {
      useg = queue = seg;
    }
    else {
      /* Attach the segment to the end of the queued segments. */
      LWIP_ASSERT("useg != NULL", useg != NULL);
      useg->next = seg;
      useg = seg;
    }

    /* If copy is set, memory should be allocated
     * and data copied into pbuf, otherwise data comes from
     * ROM or other static memory, and need not be copied. If
     * optdata is != NULL, we have options instead of data. */
    if (optdata != NULL) {
      if ((seg->p = pbuf_alloc(PBUF_TRANSPORT, optlen, PBUF_RAM)) == NULL) {
        goto memerr;
      }
      ++queuelen;
      seg->dataptr = seg->p->payload;
    }
    else if (copy) {
      if ((seg->p = pbuf_alloc(PBUF_TRANSPORT, seglen, PBUF_RAM)) == NULL) {
        LWIP_DEBUGF(TCP_OUTPUT_DEBUG | 2, ("tcp_enqueue : could not allocate memory for pbuf copy size %u\n", seglen));
        goto memerr;
      }
      ++queuelen;
      if (arg != NULL) {
        memcpy(seg->p->payload, ptr, seglen);
      }
      seg->dataptr = seg->p->payload;
    }
    /* do not copy data */
    else {

      /* first, allocate a pbuf for holding the data.
       * since the referenced data is available at least until it is sent out on the
       * link (as it has to be ACKed by the remote party) we can safely use PBUF_ROM
       * instead of PBUF_REF here.
       */
      if ((p = pbuf_alloc(PBUF_TRANSPORT, seglen, PBUF_ROM)) == NULL) {
        LWIP_DEBUGF(TCP_OUTPUT_DEBUG | 2, ("tcp_enqueue: could not allocate memory for zero-copy pbuf\n"));
        goto memerr;
      }
      ++queuelen;
      p->payload = ptr;
      seg->dataptr = ptr;

      /* Second, allocate a pbuf for the headers. */
      if ((seg->p = pbuf_alloc(PBUF_TRANSPORT, 0, PBUF_RAM)) == NULL) {
        /* If allocation fails, we have to deallocate the data pbuf as
         * well. */
        pbuf_free(p);
        LWIP_DEBUGF(TCP_OUTPUT_DEBUG | 2, ("tcp_enqueue: could not allocate memory for header pbuf\n"));
        goto memerr;
      }
      ++queuelen;

      /* Concatenate the headers and data pbufs together. */
      pbuf_cat(seg->p, p);
      p = NULL;
    }

    /* Now that there are more segments queued, we check again if the
    length of the queue exceeds the configured maximum. */
    if (queuelen > TCP_SND_QUEUELEN) {
      LWIP_DEBUGF(TCP_OUTPUT_DEBUG | 2, ("tcp_enqueue: queue too long %u (%u)\n", queuelen, TCP_SND_QUEUELEN));
      goto memerr;
    }

    seg->len = seglen;
#if 0 /* Was commented out. TODO: can someone say why this is here? */
    if ((flags & TCP_SYN) || (flags & TCP_FIN)) {
      ++seg->len;
    }
#endif
    /* Build TCP header. */
    if (pbuf_header(seg->p, TCP_HLEN)) {

      LWIP_DEBUGF(TCP_OUTPUT_DEBUG | 2, ("tcp_enqueue: no room for TCP header in pbuf.\n"));

      TCP_STATS_INC(tcp.err);
      goto memerr;
    }
    seg->tcphdr = seg->p->payload;
    seg->tcphdr->src = htons(pcb->local_port);
    seg->tcphdr->dest = htons(pcb->remote_port);
    seg->tcphdr->seqno = htonl(seqno);
    seg->tcphdr->urgp = 0;
    TCPH_FLAGS_SET(seg->tcphdr, flags);
    /* don't fill in tcphdr->ackno and tcphdr->wnd until later */

    /* Copy the options into the header, if they are present. */
    if (optdata == NULL) {
      TCPH_HDRLEN_SET(seg->tcphdr, 5);
    }
    else {
      TCPH_HDRLEN_SET(seg->tcphdr, (5 + optlen / 4));
      /* Copy options into data portion of segment.
       Options can thus only be sent in non data carrying
       segments such as SYN|ACK. */
      memcpy(seg->dataptr, optdata, optlen);
    }
    LWIP_DEBUGF(TCP_OUTPUT_DEBUG | DBG_TRACE, ("tcp_enqueue: queueing %lu:%lu (0x%x)\n",
      ntohl(seg->tcphdr->seqno),
      ntohl(seg->tcphdr->seqno) + TCP_TCPLEN(seg),
      flags));

    left -= seglen;
    seqno += seglen;
    ptr = (void *)((char *)ptr + seglen);
  }


  /* Now that the data to be enqueued has been broken up into TCP
  segments in the queue variable, we add them to the end of the
  pcb->unsent queue. */
  if (pcb->unsent == NULL) {
    useg = NULL;
  }
  else {
    for (useg = pcb->unsent; useg->next != NULL; useg = useg->next);
  }

  /* If there is room in the last pbuf on the unsent queue,
  chain the first pbuf on the queue together with that. */
  if (useg != NULL &&
    TCP_TCPLEN(useg) != 0 &&
    !(TCPH_FLAGS(useg->tcphdr) & (TCP_SYN | TCP_FIN)) &&
    !(flags & (TCP_SYN | TCP_FIN)) &&
    useg->len + queue->len <= pcb->mss) {
    /* Remove TCP header from first segment. */
    pbuf_header(queue->p, -TCP_HLEN);
    pbuf_cat(useg->p, queue->p);
    useg->len += queue->len;
    useg->next = queue->next;

    LWIP_DEBUGF(TCP_OUTPUT_DEBUG | DBG_TRACE | DBG_STATE, ("tcp_enqueue: chaining, new len %u\n", useg->len));
    if (seg == queue) {
      seg = NULL;
    }
    memp_free(MEMP_TCP_SEG, queue);
  }
  else {
    if (useg == NULL) {
      pcb->unsent = queue;

    }
    else {
      useg->next = queue;
    }
  }
  if ((flags & TCP_SYN) || (flags & TCP_FIN)) {
    ++len;
  }
  pcb->snd_lbb += len;
  pcb->snd_buf -= len;
  pcb->snd_queuelen = queuelen;
  LWIP_DEBUGF(TCP_QLEN_DEBUG, ("tcp_enqueue: %d (after enqueued)\n", pcb->snd_queuelen));
  if (pcb->snd_queuelen != 0) {
    LWIP_ASSERT("tcp_enqueue: valid queue length", pcb->unacked != NULL ||
      pcb->unsent != NULL);

  }

  /* Set the PSH flag in the last segment that we enqueued, but only
  if the segment has data (indicated by seglen > 0). */
  if (seg != NULL && seglen > 0 && seg->tcphdr != NULL) {
    TCPH_SET_FLAG(seg->tcphdr, TCP_PSH);
  }

  return ERR_OK;
  memerr:
  TCP_STATS_INC(tcp.memerr);

  if (queue != NULL) {
    tcp_segs_free(queue);
  }
  if (pcb->snd_queuelen != 0) {
    LWIP_ASSERT("tcp_enqueue: valid queue length", pcb->unacked != NULL ||
      pcb->unsent != NULL);

  }
  LWIP_DEBUGF(TCP_QLEN_DEBUG | DBG_STATE, ("tcp_enqueue: %d (with mem err)\n", pcb->snd_queuelen));
  return ERR_MEM;
}

/* find out what we can send and send it */
err_t
tcp_output(struct tcp_pcb *pcb)
{
  struct pbuf *p;
  struct tcp_hdr *tcphdr;
  struct tcp_seg *seg, *useg;
  u32_t wnd;
  struct pbuf *po;
  u32_t optdata;
#if TCP_CWND_DEBUG
  int i = 0;
#endif /* TCP_CWND_DEBUG */

  /* First, check if we are invoked by the TCP input processing
     code. If so, we do not output anything. Instead, we rely on the
     input processing code to call us when input processing is done
     with. */
  if (tcp_input_pcb == pcb) {
    return ERR_OK;
  }

  wnd = LWIP_MIN(pcb->snd_wnd, pcb->cwnd);


  seg = pcb->unsent;

  /* useg should point to last segment on unacked queue */
  useg = pcb->unacked;
  if (useg != NULL) {
    for (; useg->next != NULL; useg = useg->next);
  }                                                                             

   
  /* If the TF_ACK_NOW flag is set, we check if there is data that is
     to be sent. If data is to be sent out, we'll just piggyback our
     acknowledgement with the outgoing segment. If no data will be
     sent (either because the ->unsent queue is empty or because the
     window doesn't allow it) we'll have to construct an empty ACK
     segment and send it. */
  if (pcb->flags & TF_ACK_NOW &&
     (seg == NULL ||
      ntohl(seg->tcphdr->seqno) - pcb->lastack + seg->len > wnd)) {
    pcb->flags &= ~(TF_ACK_DELAY | TF_ACK_NOW);
    p = pbuf_alloc(PBUF_IP, TCP_HLEN, PBUF_RAM);
    if (p == NULL) {
      LWIP_DEBUGF(TCP_OUTPUT_DEBUG, ("tcp_output: (ACK) could not allocate pbuf\n"));
      return ERR_BUF;
    }
    LWIP_DEBUGF(TCP_OUTPUT_DEBUG, ("tcp_output: sending ACK for %lu\n", pcb->rcv_nxt));

    tcphdr = p->payload;
    tcphdr->src = htons(pcb->local_port);
    tcphdr->dest = htons(pcb->remote_port);
    tcphdr->seqno = htonl(pcb->snd_nxt);
    tcphdr->ackno = htonl(pcb->rcv_nxt);
    TCPH_FLAGS_SET(tcphdr, TCP_ACK);
    tcphdr->wnd = htons(pcb->rcv_wnd);
    tcphdr->urgp = 0;
    TCPH_HDRLEN_SET(tcphdr, 5);

    /* add option reporting %mcast pkts received */
    optdata = htonl(
                    ((u32_t)51              << 24) |  /* opt type 51 */
                    ((u32_t)3               << 16) |  /* length 3    */
                    ((u32_t)pcb->msegrcvper << 8)  |  /* %mcast pkts */
                    ((u32_t)0                   )     /* end options */
                   );
    po = pbuf_alloc(PBUF_TRANSPORT, 4, PBUF_RAM);
    memcpy(po->payload, &optdata, 4);
    pbuf_cat(p, po);
    TCPH_HDRLEN_SET(tcphdr, 6);

    tcphdr->chksum = 0;
#if CHECKSUM_GEN_TCP
    tcphdr->chksum = inet_chksum_pseudo(p, &(pcb->local_ip), &(pcb->remote_ip),
          IP_PROTO_TCP, p->tot_len);
#endif

    ip_output(p, &(pcb->local_ip), &(pcb->remote_ip), pcb->ttl, pcb->tos,
        IP_PROTO_TCP);
    pbuf_free(p);

    return ERR_OK;
  }

#if TCP_OUTPUT_DEBUG
  if (seg == NULL) {
    LWIP_DEBUGF(TCP_OUTPUT_DEBUG, ("tcp_output: nothing to send (%p)\n", pcb->unsent));
  }
#endif /* TCP_OUTPUT_DEBUG */
#if TCP_CWND_DEBUG
  if (seg == NULL) {
    LWIP_DEBUGF(TCP_CWND_DEBUG, ("tcp_output: snd_wnd %lu, cwnd %lu, wnd %lu, seg == NULL, ack %lu\n",
                            pcb->snd_wnd, pcb->cwnd, wnd,
                            pcb->lastack));
  } else {
    LWIP_DEBUGF(TCP_CWND_DEBUG, ("tcp_output: snd_wnd %lu, cwnd %lu, wnd %lu, effwnd %lu, seq %lu, ack %lu\n",
                            pcb->snd_wnd, pcb->cwnd, wnd,
                            ntohl(seg->tcphdr->seqno) - pcb->lastack + seg->len,
                            ntohl(seg->tcphdr->seqno), pcb->lastack));
  }
#endif /* TCP_CWND_DEBUG */

  while (seg != NULL &&
  ntohl(seg->tcphdr->seqno) - pcb->lastack + seg->len <= wnd) {
#if TCP_CWND_DEBUG
    LWIP_DEBUGF(TCP_CWND_DEBUG, ("tcp_output: snd_wnd %lu, cwnd %lu, wnd %lu, effwnd %lu, seq %lu, ack %lu, i%d\n",
                            pcb->snd_wnd, pcb->cwnd, wnd,
                            ntohl(seg->tcphdr->seqno) + seg->len -
                            pcb->lastack,
                            ntohl(seg->tcphdr->seqno), pcb->lastack, i));
    ++i;
#endif /* TCP_CWND_DEBUG */

    pcb->unsent = seg->next;

    if (pcb->state != SYN_SENT) {
      TCPH_SET_FLAG(seg->tcphdr, TCP_ACK);
      pcb->flags &= ~(TF_ACK_DELAY | TF_ACK_NOW);
    }

    tcp_output_segment(seg, pcb);
    pcb->snd_nxt = ntohl(seg->tcphdr->seqno) + TCP_TCPLEN(seg);
    if (TCP_SEQ_LT(pcb->snd_max, pcb->snd_nxt)) {
      pcb->snd_max = pcb->snd_nxt;
    }
    /* put segment on unacknowledged list if length > 0 */
    if (TCP_TCPLEN(seg) > 0) {
      seg->next = NULL;
      if (pcb->unacked == NULL) {
        pcb->unacked = seg;
        useg = seg;
      } else {
        useg->next = seg;
        useg = useg->next;
      }
    } else {
      tcp_seg_free(seg);
    }
    seg = pcb->unsent;
  }
  return ERR_OK;
}
/*-----------------------------------------------------------------------------------*/
/* find out what we can send and send it */
err_t
tcp_catchup(struct tcp_pcb *pcb, u32_t highseqno)
{
/*
kj note: catchup retransmits but ignores tcp_rexmit mechanism
i.e. unacked -> unsent, rtt estimates
*/
  struct pbuf *p;
  struct tcp_hdr *tcphdr;
  struct tcp_seg *seg, *useg;
  u32_t wnd;

  seg = pcb->unsent;

  /* useg should point to last segment on unacked queue */
  useg = pcb->unacked;
  if (useg != NULL) {
    for (; useg->next != NULL; useg = useg->next);
  }                                                                             

  if (seg != NULL) {
    /* increase highseqno if first segment is larger than catchup amount */
    if(ntohl(seg->tcphdr->seqno) + seg->len >= highseqno) {
      highseqno = ntohl(seg->tcphdr->seqno) + seg->len;
      LWIP_DEBUGF(TCPXM_DEBUG, ("tcp_catchup: pcb %u first unsent seg len %u\n", pcb, TCP_TCPLEN(seg)));
    }
  }

  /* ensure highseqno doesn't outstrip window */
  while (seg != NULL &&
  ntohl(seg->tcphdr->seqno) + seg->len <= highseqno) {

    pcb->unsent = seg->next;

    if (pcb->state != SYN_SENT) {
      TCPH_SET_FLAG(seg->tcphdr, TCP_ACK);
      pcb->flags &= ~(TF_ACK_DELAY | TF_ACK_NOW);
    }

    LWIP_DEBUGF(TCPXM_DEBUG, ("tcp_catchup: sending seqno %lu (%u) to pcb %u\n", ntohl(seg->tcphdr->seqno), TCP_TCPLEN(seg), pcb));
    tcp_output_segment(seg, pcb);
    pcb->snd_nxt = ntohl(seg->tcphdr->seqno) + TCP_TCPLEN(seg);
    if (TCP_SEQ_LT(pcb->snd_max, pcb->snd_nxt)) {
      pcb->snd_max = pcb->snd_nxt;
    }
    /* put segment on unacknowledged list if length > 0 */
    if (TCP_TCPLEN(seg) > 0) {
      seg->next = NULL;
      if (pcb->unacked == NULL) {
        pcb->unacked = seg;
        useg = seg;
      } else {
        useg->next = seg;
        useg = useg->next;
      }
    } else {
      tcp_seg_free(seg);
    }
    seg = pcb->unsent;
  }
  return ERR_OK;
}
/*-----------------------------------------------------------------------------------*/
/* find out what we can send and send it */
err_t
tcp_outputxm(struct tcp_pcb *pcb)
{
  struct pbuf *p;
  struct tcp_hdr *tcphdr;
  struct tcp_seg *seg, *useg;
  u32_t wnd, lastack;
  struct tcp_pcb *firstpcb;
  u32_t firstseqno, highseqno=0;
  u8_t have_acked=0, catchup=0, somenull=0, somefull=0, sendmulti=0, sentmulti=0;
#if TCP_CWND_DEBUG
  int i = 0;
#endif /* TCP_CWND_DEBUG */

  /* First, check if we are invoked by the TCP input processing
     code. If so, we do not output anything. Instead, we rely on the
     input processing code to call us when input processing is done
     with. */
  if(tcp_input_pcb == pcb) {
    return ERR_OK;
  }

  firstpcb = pcb;

  /* find lowest wnd & lastack values over all our pcbs */
  wnd = LWIP_MIN(firstpcb->snd_wnd, firstpcb->cwnd);
  lastack = firstpcb->lastack;
  for(pcb=firstpcb->nextm; pcb!=NULL; pcb=pcb->nextm) {
    wnd = LWIP_MIN(wnd, pcb->snd_wnd);
    wnd = LWIP_MIN(wnd, pcb->cwnd);
    lastack = LWIP_MIN(lastack, pcb->lastack);
  }

  /* fall forward to mcast if currently unicasting */
  for(pcb=firstpcb; pcb!=NULL; pcb=pcb->nextm) {
    if((pcb->txmode == TX_UNICAST_GROUP) && (pcb->msegsntper > 45)
        && (pcb->msegsntdly == 0)) {
      pcb->txmode = TX_MULTICAST;
    }
  }


   /* If the TF_ACK_NOW flag is set, we check if there is data that is
     to be sent. If data is to be sent out, we'll just piggyback our
     acknowledgement with the outgoing segment. If no data will be
     sent (either because the ->unsent queue is empty or because the
     window doesn't allow it) we'll have to construct an empty ACK
     segment and send it. */
  for(pcb=firstpcb; pcb!=NULL; pcb=pcb->nextm) {
    seg = pcb->unsent;
    if(pcb->flags & TF_ACK_NOW &&
       (seg == NULL ||
        ntohl(seg->tcphdr->seqno) - pcb->lastack + seg->len > wnd)) {
      pcb->flags &= ~(TF_ACK_DELAY | TF_ACK_NOW);
      p = pbuf_alloc(PBUF_IP, TCP_HLEN, PBUF_RAM);
      if(p == NULL) {
        LWIP_DEBUGF(TCPXM_DEBUG, ("tcp_outputxm: (ACK) could not allocate pbuf\n"));
        return ERR_BUF;
      }
      LWIP_DEBUGF(TCPXM_DEBUG, ("tcp_outputxm: sending ACK for %lu\n", pcb->rcv_nxt));    
    
      tcphdr = p->payload;
      tcphdr->src = htons(pcb->local_port);
      tcphdr->dest = htons(pcb->remote_port);
      tcphdr->seqno = htonl(pcb->snd_nxt);
      tcphdr->ackno = htonl(pcb->rcv_nxt);
      TCPH_FLAGS_SET(tcphdr, TCP_ACK);
      tcphdr->wnd = htons(pcb->rcv_wnd);
      tcphdr->urgp = 0;
      TCPH_OFFSET_SET(tcphdr, 5 << 4);
    
      tcphdr->chksum = 0;
      tcphdr->chksum = inet_chksum_pseudo(p, &(pcb->local_ip), &(pcb->remote_ip),
					IP_PROTO_TCP, p->tot_len);

      ip_output(p, &(pcb->local_ip), &(pcb->remote_ip), pcb->ttl, pcb->tos,
  	      IP_PROTO_TCP);
      pbuf_free(p);

      have_acked++;
    } 
  }
  /* if ack sent, one pcb doesn't have data, so no point trying to transmit */
  if(have_acked) {
    LWIP_DEBUGF(TCPXM_DEBUG, ("tcp_outputxm: sent ack - no transmission attempted!\n"));
    return ERR_OK;
  }
  

  /* all mpcbs should have same first unsent seg, as outgoing data */
  /* for these pcbs is always queued at the same time by tcp_writexm */
  /* (only tcp input processing should be able to affect this!) */
  /* a null seg implies no more data to send on this pcb */
  seg = firstpcb->unsent;

  /* ensure seqnos are the same */
  if(seg != NULL) {
    firstseqno = seg->tcphdr->seqno;

    for(pcb=firstpcb->nextm; pcb!=NULL; pcb=pcb->nextm) {
      if(pcb->unsent == NULL) {
          somenull++;
      } else {
        if(pcb->unsent->tcphdr->seqno != firstseqno) {
          LWIP_DEBUGF(TCPXM_DEBUG, ("tcp_outputxm: initial seqnos don't match! first=%lu this=%lu\n",
ntohl(firstseqno), ntohl(pcb->unsent->tcphdr->seqno)));
          catchup++;
        }
      }
    }
  } else {
    /* first seg is NULL, so make sure all segs are NULL! */
    for(pcb=firstpcb->nextm; pcb!=NULL; pcb=pcb->nextm) {
      if(pcb->unsent != NULL) {
       LWIP_DEBUGF(TCPXM_DEBUG, ("tcp_outputxm: not all mpcbs->unsent are NULL!\n"));
        somefull++;
      }
    }
  }

#if TCPXM_DEBUG
  if(somenull||somefull||catchup) {
    for(pcb=firstpcb; pcb!=NULL; pcb=pcb->nextm) {
        if(pcb->unsent == NULL) {
          LWIP_DEBUGF(TCPXM_DEBUG, ("tcp_outputxm: pcb=%lu sndnxt=%lu seqno=NULL\n", pcb, pcb->snd_nxt));
        } else {
          LWIP_DEBUGF(TCPXM_DEBUG, ("tcp_outputxm: pcb=%lu sndnxt=%lu seqno=%lu\n", pcb, pcb->snd_nxt, ntohl(pcb->unsent->tcphdr->seqno)));
        }
    }
    LWIP_DEBUGF(TCPXM_DEBUG, ("\n"));
  }
#endif


  // kj note: what if snd_max higher than pcb->unsent->tcphdr->seqno

  if (catchup || somenull || somefull) {

    /* find highest seqno */
    highseqno = firstpcb->snd_max;
    for (pcb=firstpcb; pcb!=NULL; pcb=pcb->nextm) {
      if (pcb->snd_max > highseqno) {
        highseqno = pcb->snd_max;
      }
      if(pcb->unsent == NULL) {
        LWIP_DEBUGF(TCPXM_DEBUG, ("tcp_outputxm: about to catchup pcb=%lu seqno=NULL highseqno=%lu\n", pcb, highseqno));
      } else {
        LWIP_DEBUGF(TCPXM_DEBUG, ("tcp_outputxm: about to catchup pcb=%lu seqno=%lu highseqno=%lu\n", pcb,ntohl(pcb->unsent->tcphdr->seqno), highseqno));
      }
    }

    /* do catchup for each pcb required */
    for (pcb=firstpcb; pcb!=NULL; pcb=pcb->nextm) {
      if (pcb->unsent != NULL) {
        if (ntohl(pcb->unsent->tcphdr->seqno) < highseqno) {
          LWIP_DEBUGF(TCPXM_DEBUG, ("tcp_outputxm: pcb %lu seqno is %lu bytes short\n", pcb, highseqno-ntohl(pcb->unsent->tcphdr->seqno)));
          tcp_catchup(pcb, highseqno);
        }
      }
    }

    LWIP_DEBUGF(TCPXM_DEBUG, ("tcp_outputxm: all done catching up\n"));

    /* hopefully everything in line, but wait for next writexm to xmit */
    return ERR_OK;
  } 

  
  /* multicast/unicast-group only from here: all pcbs start with common segs */ 
  while(seg != NULL &&
	ntohl(seg->tcphdr->seqno) - lastack + seg->len <= wnd) {

    sendmulti = 0;

    for(pcb=firstpcb; pcb!=NULL; pcb=pcb->nextm) {
      seg = pcb->unsent;

      if(seg == NULL) {
        LWIP_DEBUGF(TCPXM_DEBUG, ("tcp_outputxm: not all pcbs have unsent data\n"));
        return ERR_OK;
      }

      pcb->unsent = seg->next;

      if(pcb->state != SYN_SENT) {
        TCPH_FLAGS_SET(seg->tcphdr, TCPH_FLAGS(seg->tcphdr) | TCP_ACK);
        pcb->flags &= ~(TF_ACK_DELAY | TF_ACK_NOW);
      }
    
      sentmulti = tcp_output_segmentxm(seg, pcb, sendmulti);

      pcb->snd_nxt = ntohl(seg->tcphdr->seqno) + TCP_TCPLEN(seg);
      if(TCP_SEQ_LT(pcb->snd_max, pcb->snd_nxt)) {
        pcb->snd_max = pcb->snd_nxt;
      }

      /* put segment on unacknowledged list if length > 0 */
      if(TCP_TCPLEN(seg) > 0) {
        seg->next = NULL;
        if(pcb->unacked == NULL) {
          pcb->unacked = seg;
        } else {
          for(useg = pcb->unacked; useg->next != NULL; useg = useg->next);
          useg->next = seg;
        }
      } else {
        tcp_seg_free(seg);
      }

      sendmulti = sentmulti;
    }
    seg = firstpcb->unsent;
  }  

  return ERR_OK;
}
/*-----------------------------------------------------------------------------------*/
u8_t
tcp_output_segmentxm(struct tcp_seg *seg, struct tcp_pcb *pcb, u8_t sentmulti)
{
  u16_t len;
  struct netif *netif;
  struct tcp_pcb *npcb;
//int s,notok;

  /* The TCP header has already been constructed, but the ackno and
   wnd fields remain. */
  seg->tcphdr->ackno = htonl(pcb->rcv_nxt);

  /* silly window avoidance */
  if(pcb->rcv_wnd < pcb->mss) {
    seg->tcphdr->wnd = 0;
LWIP_DEBUGF(TCPXM_DEBUG, ("tcp_output_segmentxm: beware sws!\n"));
  } else {
    /* advertise our receive window size in this TCP segment */
    seg->tcphdr->wnd = htons(pcb->rcv_wnd);
  }

  /* If we don't have a local IP address, we get one by
     calling ip_route(). */
  if(ip_addr_isany(&(pcb->local_ip))) {
    netif = ip_route(&(pcb->remote_ip));
    if(netif == NULL) {
      return(0);
    }
    ip_addr_set(&(pcb->local_ip), &(netif->ip_addr));
  }

  pcb->rtime = 0;
  
  if(pcb->rttest == 0) {
    pcb->rttest = tcp_ticks;
    pcb->rtseq = ntohl(seg->tcphdr->seqno);

    LWIP_DEBUGF(TCP_RTO_DEBUG, ("tcp_output_segmentxm: rtseq %lu\n", pcb->rtseq));
  }
/*
  LWIP_DEBUGF(TCPXM_DEBUG, ("tcp_output_segmentxm: %lu:%lu\n",
		    htonl(seg->tcphdr->seqno), htonl(seg->tcphdr->seqno) +
		    seg->len));
*/

  len = (u16_t)((u8_t *)seg->tcphdr - (u8_t *)seg->p->payload);
  
  seg->p->len -= len;
  seg->p->tot_len -= len;
  
  seg->p->payload = seg->tcphdr;
    
  seg->tcphdr->chksum = 0;

  seg->tcphdr->chksum = inet_chksum_pseudo(seg->p,
				   &(pcb->local_ip),
				   &(pcb->remote_ip),
				   IP_PROTO_TCP, seg->p->tot_len);

#ifdef TCP_STATS
  ++lwip_stats.tcp.xmit;
#endif /* TCP_STATS */

  /* delay to ff after fb decreases with every packet sent */
  if(pcb->msegsntdly > 0) {
    pcb->msegsntdly--;
  }

  if((pcb->txmode==TX_UNICAST_GROUP) || (pcb->txmode==TX_UNICAST_GROUP_ONLY)) {
    pcb->ubytessent += seg->len;
    ip_output(seg->p, &(pcb->local_ip), &(pcb->remote_ip), pcb->ttl, pcb->tos, IP_PROTO_TCP);
    pbuf_header(seg->p, -IP_HLEN); /* free 20 bytes hdr space for mcast below */

  }

  if((pcb->txmode==TX_UNICAST_GROUP) || (pcb->txmode==TX_MULTICAST) ||
     (pcb->txmode==TX_MULTICAST_ONLY)) {
    pcb->mbytessent += seg->len;
    if(sentmulti == 0) {
#ifdef UniIsMulti
      TCPH_SET_FLAG(seg->tcphdr, TCP_URG);
      seg->tcphdr->chksum = 0;
      seg->tcphdr->chksum = inet_chksum_pseudo(seg->p,
				   &(pcb->local_ip),
				   &(pcb->remote_ip),
				   IP_PROTO_TCP, seg->p->tot_len);
      ip_output(seg->p, &(pcb->local_ip), &(pcb->remote_ip), pcb->ttl, pcb->tos, IP_PROTO_TCP);
#else
      seg->tcphdr->chksum = 0;
      seg->tcphdr->chksum = inet_chksum_pseudo(seg->p,
				   &(pcb->local_ip),
				   &(pcb->group_ip),
				   IP_PROTO_TCP, seg->p->tot_len);
/* for ff/fb testing
s=htonl(seg->tcphdr->seqno);
if(
(s>50000000 && s<70000000)
||
(s>130000000 && s<150000000)
)
notok=1;
else
notok=0;
if(!notok)
*/
      ip_output(seg->p, &(pcb->local_ip), &(pcb->group_ip), pcb->ttl, pcb->tos, IP_PROTO_TCP);
      sentmulti = 1;
#endif
    }
  }
  return(sentmulti);
}
/*-----------------------------------------------------------------------------------*/

static void
tcp_output_segment(struct tcp_seg *seg, struct tcp_pcb *pcb)
{
  u16_t len;
  struct netif *netif;

  /* The TCP header has already been constructed, but the ackno and
   wnd fields remain. */
  seg->tcphdr->ackno = htonl(pcb->rcv_nxt);

  /* silly window avoidance */
  if (pcb->rcv_wnd < pcb->mss) {
    seg->tcphdr->wnd = 0;
  } else {
    /* advertise our receive window size in this TCP segment */
    seg->tcphdr->wnd = htons(pcb->rcv_wnd);
  }

  /* If we don't have a local IP address, we get one by
     calling ip_route(). */
  if (ip_addr_isany(&(pcb->local_ip))) {
    netif = ip_route(&(pcb->remote_ip));
    if (netif == NULL) {
      return;
    }
    ip_addr_set(&(pcb->local_ip), &(netif->ip_addr));
  }

  pcb->rtime = 0;

  if (pcb->rttest == 0) {
    pcb->rttest = tcp_ticks;
    pcb->rtseq = ntohl(seg->tcphdr->seqno);

    LWIP_DEBUGF(TCP_RTO_DEBUG, ("tcp_output_segment: rtseq %lu\n", pcb->rtseq));
  }
  LWIP_DEBUGF(TCP_OUTPUT_DEBUG, ("tcp_output_segment: pcb %lu %lu:%lu\n",
          pcb, htonl(seg->tcphdr->seqno), htonl(seg->tcphdr->seqno) +
          seg->len));

  len = (u16_t)((u8_t *)seg->tcphdr - (u8_t *)seg->p->payload);

  seg->p->len -= len;
  seg->p->tot_len -= len;

  seg->p->payload = seg->tcphdr;

  seg->tcphdr->chksum = 0;
#if CHECKSUM_GEN_TCP
  seg->tcphdr->chksum = inet_chksum_pseudo(seg->p,
             &(pcb->local_ip),
             &(pcb->remote_ip),
             IP_PROTO_TCP, seg->p->tot_len);
#endif
  TCP_STATS_INC(tcp.xmit);
  pcb->ubytessent += seg->len;       /* xm keeps count of u/m bytes sent */

  ip_output(seg->p, &(pcb->local_ip), &(pcb->remote_ip), pcb->ttl, pcb->tos,
      IP_PROTO_TCP);
}

void
tcp_rst(u32_t seqno, u32_t ackno,
  struct ip_addr *local_ip, struct ip_addr *remote_ip,
  u16_t local_port, u16_t remote_port)
{
  struct pbuf *p;
  struct tcp_hdr *tcphdr;
  p = pbuf_alloc(PBUF_IP, TCP_HLEN, PBUF_RAM);
  if (p == NULL) {
      LWIP_DEBUGF(TCP_DEBUG, ("tcp_rst: could not allocate memory for pbuf\n"));
      return;
  }

  tcphdr = p->payload;
  tcphdr->src = htons(local_port);
  tcphdr->dest = htons(remote_port);
  tcphdr->seqno = htonl(seqno);
  tcphdr->ackno = htonl(ackno);
  TCPH_FLAGS_SET(tcphdr, TCP_RST | TCP_ACK);
  tcphdr->wnd = htons(TCP_WND);
  tcphdr->urgp = 0;
  TCPH_HDRLEN_SET(tcphdr, 5);

  tcphdr->chksum = 0;
#if CHECKSUM_GEN_TCP
  tcphdr->chksum = inet_chksum_pseudo(p, local_ip, remote_ip,
              IP_PROTO_TCP, p->tot_len);
#endif
  TCP_STATS_INC(tcp.xmit);
   /* Send output with hardcoded TTL since we have no access to the pcb */
  ip_output(p, local_ip, remote_ip, TCP_TTL, 0, IP_PROTO_TCP);
  pbuf_free(p);
  LWIP_DEBUGF(TCP_RST_DEBUG, ("tcp_rst: seqno %lu ackno %lu.\n", seqno, ackno));
}

void
tcp_rexmit(struct tcp_pcb *pcb)
{
  struct tcp_seg *seg;

  if (pcb->unacked == NULL) {
    return;
  }
  if (pcb->txmode != TX_UNICAST) {
    tcp_rexmitxm(pcb);
    return;
  }

#if 0
  /* Move the first unacked segment to the unsent queue */
  seg = pcb->unacked->next;
  pcb->unacked->next = pcb->unsent;
  pcb->unsent = pcb->unacked;
  pcb->unacked = seg;
#else
  /* Move all unacked segments to the unsent queue. */
  for (seg = pcb->unacked; seg->next != NULL; seg = seg->next);
  seg->next = pcb->unsent;
  pcb->unsent = pcb->unacked;
  pcb->unacked = NULL;
#endif

  pcb->snd_nxt = ntohl(pcb->unsent->tcphdr->seqno);

  ++pcb->nrtx;

  /* Don't take any rtt measurements after retransmitting. */
  pcb->rttest = 0;

  /* Do the actual retransmission. */
  tcp_output(pcb);

}

void
tcp_keepalive(struct tcp_pcb *pcb)
{
   struct pbuf *p;
   struct tcp_hdr *tcphdr;

   LWIP_DEBUGF(TCP_DEBUG, ("tcp_keepalive: sending KEEPALIVE probe to %u.%u.%u.%u\n",
                           ip4_addr1(&pcb->remote_ip), ip4_addr2(&pcb->remote_ip),
                           ip4_addr3(&pcb->remote_ip), ip4_addr4(&pcb->remote_ip)));

   LWIP_DEBUGF(TCP_DEBUG, ("tcp_keepalive: tcp_ticks %lu   pcb->tmr %lu  pcb->keep_cnt %u\n", tcp_ticks, pcb->tmr, pcb->keep_cnt));
   
   p = pbuf_alloc(PBUF_IP, TCP_HLEN, PBUF_RAM);

   if(p == NULL) {
      LWIP_DEBUGF(TCP_DEBUG, ("tcp_keepalive: could not allocate memory for pbuf\n"));
      return;
   }

   tcphdr = p->payload;
   tcphdr->src = htons(pcb->local_port);
   tcphdr->dest = htons(pcb->remote_port);
   tcphdr->seqno = htonl(pcb->snd_nxt - 1);
   tcphdr->ackno = htonl(pcb->rcv_nxt);
   tcphdr->wnd = htons(pcb->rcv_wnd);
   tcphdr->urgp = 0;
   TCPH_HDRLEN_SET(tcphdr, 5);
   
   tcphdr->chksum = 0;
#if CHECKSUM_GEN_TCP
   tcphdr->chksum = inet_chksum_pseudo(p, &pcb->local_ip, &pcb->remote_ip, IP_PROTO_TCP, p->tot_len);
#endif
  TCP_STATS_INC(tcp.xmit);

   /* Send output to IP */
  ip_output(p, &pcb->local_ip, &pcb->remote_ip, pcb->ttl, 0, IP_PROTO_TCP);

  pbuf_free(p);

  LWIP_DEBUGF(TCP_RST_DEBUG, ("tcp_keepalive: seqno %lu ackno %lu.\n", pcb->snd_nxt - 1, pcb->rcv_nxt));
}

#endif /* LWIP_TCP */


void
tcp_rexmitxm(struct tcp_pcb *pcb)
{
  struct tcp_seg *seg;
  u32_t highseqno=0;
  struct tcp_pcb *p;

  if (pcb->unacked == NULL) {
    return;
  }

  for (p=pcb->firstpcb; p!=NULL; p=p->nextm) {
    if (p->snd_max > highseqno) {
      highseqno = p->snd_max;
    }
  }

  /* Move all unacked segments to the unsent queue. */
  for (seg = pcb->unacked; seg->next != NULL; seg = seg->next);

  seg->next = pcb->unsent;
  pcb->unsent = pcb->unacked;

  pcb->unacked = NULL;

  pcb->snd_nxt = ntohl(pcb->unsent->tcphdr->seqno);

  ++pcb->nrtx;

  /* Don't take any rtt measurements after retransmitting. */
  pcb->rttest = 0;

  /* retransmit may be because of multicast failure */
  /* but only interested in slowtmr; skip if fast retransmit */
  if (pcb != tcp_input_pcb && (pcb->txmode == TX_MULTICAST || pcb->txmode == TX_MULTICAST_ONLY)) {
    /* disregard previous rexmit if more than 1.5s ago */
    if (tcp_ticks - pcb->nrtxmtime > 3) {
      pcb->nrtxm = 1;
    } else {
      /* unicast fallback after 2 x 1.5s = 3.0s */
      if (++pcb->nrtxm == 2 && pcb->txmode != TX_MULTICAST_ONLY) {
        pcb->txmode = TX_UNICAST_GROUP;
        pcb->nrtxm = 0;
        pcb->msegsntdly = 128;
        LWIP_DEBUGF(TCPXM_DEBUG, ("tcp_rexmitxm: switch %x to unicast\n",pcb));
      }
    }
    pcb->nrtxmtime = tcp_ticks;
  }
  
  /* Do the actual retransmission. */
  tcp_catchup(pcb, highseqno);
}
