/*
 * mcpd.c
 *
 * Multicast Copy Daemon
 * Karl Jeacle
 * University of Cambridge
 *
 * karl.jeacle@cl.cam.ac.uk
 */

#include <stdio.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/signal.h>

#include "tcpxm.h"
#include "mcp.h"

#ifndef TCPXM
netconn_waitclose(){}
#endif

int
main(int argc, char **argv)
{
  struct mysockaddr_in cli_addr, serv_addr;
  struct mcpcntl_s mcpcntl;
  struct netconn *conn, *newconn;
  struct netbuf *buf;
  char ch, data[BLOCKSIZE], msg[80], dirname[64];
  int vflag=0, dflag=0, zflag=0, pflag=MCPDTCPPORT, kflag=0, Kflag=0, tflag=0;
  int cmd, len, fd, r, pid=0;
  int s, newsock, s2, newsock2, clilen, flag_on=1;
  unsigned long bytesrecd=0, ubytesrcvd=0, mbytesrcvd=0, ubytesrcvd_early=0;
  int mcp_dstport;
  int checksock;
  void handler();


  while((ch = getopt(argc, argv, "d:hkKp:t:vz")) != -1) {
    switch (ch) {
      case 'd':
        dflag = 1;
        strcpy(dirname, optarg);
        break;
      case 'h':
        helpusage();
        break;
      case 'p':
        pflag = atoi(optarg);
        break;
      case 'k':
        kflag = 1;
        break;
      case 'K':
        Kflag = 1;
        break;
      case 't':
        tflag = atoi(optarg);
        break;
      case 'v':
        vflag = 1;
        break;
      case 'z':
        zflag = 1;
        break;
      case '?':
      default:
        usage();
    }
  }
  argc -= optind;
  argv += optind;

  if(argc!=0) {
    usage();
  }

  /* kill all other mcpds if required */
  if(kflag) {
    killmcpds(vflag);
  }
  if(Kflag) {
    killmcpds(vflag);
    exit(0);
  }

  /* try to chdir if -d specified */
  if(dflag) {
    if(chdir(dirname) < 0) {
      printf("couldn't change dir to %s\n", dirname);
      exit(1);
    }
  }

  /* zombie code was here; moved to parent after fork to avoid */
  /* loss of tapif ifconfig return values caused by no sigchld */

  if(!tflag) {
    /* socket for incoming TCP control connection */
    if((s = socket(PF_INET, SOCK_STREAM, 0)) < 0){
      perror("can't create control socket");
      exit(1);
    }

    /* set reuse port to on to allow multiple binds per host */
    if ((setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &flag_on,
         sizeof(flag_on))) < 0) {
      perror("setsockopt() failed on so_reuseport");
      exit(1);
    }

    /* bind to pflag/MCPDTCPPORT on any interface */
    bzero((char *) &serv_addr, sizeof(serv_addr));
    serv_addr.sin_family      = PF_INET;
    serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    serv_addr.sin_port        = htons(pflag);

    if(bind(s, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0){
      perror("can't bind local address");
      exit(1);
    }

    if(listen(s, 10) <0) {
      perror("listen failed");
      exit(1);
    }

    if(vflag)
      printf("tcp control bound and listening on port %d\n", pflag);
  }

  /* close stdout/stderr so remote ssh detaches */
  /* (closing stdin causes sigpipe in freebsd)  */
  if(!vflag) {
    close(STDOUT_FILENO);
    close(STDERR_FILENO);
  }

  /* parent loops forever forking children on incoming requests */
  do {

    if(!tflag) {
      clilen = sizeof(cli_addr);
      if((newsock = accept(s, (struct sockaddr *) &cli_addr, &clilen))<0){
        perror("server: control accept error");
        exit(1);
      }

      /* test mode; want UDP 2222; kill last child and other mcpds */
      if(kflag) {
        if(pid) {
          if(vflag)
            printf("killing last child pid=%d\n", pid);
          kill(pid, SIGKILL); /* last child */
        }
        killmcpds(vflag); // still need this after earlier invocation?
      }
    }

    if(!tflag) {
      pid = fork();
    }

    if(pid == 0) { /* child */

      if(!tflag) {
        if(vflag)
          printf("tcp control connected\n");

        /* receive control info from client */
        recv_control(newsock, &mcpcntl); /* srcport  */
        recv_control(newsock, &mcpcntl); /* dstport  */
        recv_control(newsock, &mcpcntl); /* filesize */
        recv_control(newsock, &mcpcntl); /* filename */
        recv_control(newsock, &mcpcntl); /* protocol */
        mcpcntl.status = MCP_NOTOK;

        if(vflag) {
          switch(mcpcntl.protocol) {
            case MCP_TCP:
              printf("incoming MultiTCP data connection\n");
              break;
            case MCP_TCPXM:
              printf("incoming tcp-xm data connection\n");
              break;
            default:
              mcpcntl.status = MCP_FAIL;
          }
        }

        /* negotiate dstport with client */
        while(mcpcntl.status == MCP_NOTOK) {

              /* check if client-suggested port is available */
              if(mcp_dstport != mcpcntl.dstport) {
                mcp_dstport = getport(mcpcntl.protocol, mcpcntl.dstport, &checksock);
                if(vflag)
                  printf("client wants dstport %d, server offering %d\n",
                    mcpcntl.dstport, mcp_dstport);
              }

              /* send back port available to us */
              send_control(newsock, MCP_DSTPORT, mcp_dstport);

              /* wait for client to say this port ok */
              recv_control(newsock, &mcpcntl); /* dstport or status*/
        }

        if(mcpcntl.status == MCP_FAIL) {
          if(vflag)
            printf("port negotiation failed - exiting\n");
          exit(1);
        }

        /* open file for writing - notify client if fail */
        if(zflag) {
          fd = open("/dev/null", O_WRONLY, 0644);
        } else {
          fd = open(mcpcntl.filename, O_CREAT|O_TRUNC|O_WRONLY, 0644);
        }
        if(fd < 0 ) {
          send_control(newsock, MCP_STATUS, MCP_NOTOK);
          send_control(newsock, MCP_MSG, strerror(errno));
          perror("open failed");
          exit(1);
        }

        if(vflag) {
          printf("receiving file %s = %d bytes\n", 
            mcpcntl.filename, mcpcntl.bytecount);
          printf("data ports: cli=%d srv=%d\n",
            mcpcntl.srcport, mcp_dstport);
        }

        close(checksock);

      } else {
          fd = open("/dev/null", O_WRONLY, 0644);
          if(fd < 0 ) {
            perror("open failed");
            exit(1);
          }
      }

      if(tflag || mcpcntl.protocol == MCP_TCPXM) {
        /* init lwip stack */
        tcpxm_init(tflag, mcp_dstport, mcpcntl.srcport);

        /* bind to tcpxm local port 7 and listen for connections */
        conn = netconn_new(NETCONN_TCP);
        netconn_bind(conn, NULL, 7);

        netconn_listen(conn);
        if(vflag)
          printf("tcp-xm data listening\n");

        if(!tflag) {
          /* tell client that we are ready to go */
          if(vflag)
            printf("sending ready to go\n");
          send_control(newsock, MCP_STATUS, MCP_OK);
        }

        newconn = netconn_accept(conn);
        if(vflag)
          printf("tcp-xm data accepted\n");

        /* read data while connection open */
        buf = netconn_recv(newconn);

        do {
          bzero(data, sizeof(data));
          netbuf_copy(buf, data, sizeof(data));
          len = netbuf_len(buf);
          bytesrecd += len;

          if((r = write(fd, data, len)) < len) {
            sprintf(msg, "wrote just %d of %d bytes to file", r, len);
	    perror(msg);
          }

          netbuf_delete(buf);

        } while((buf = netconn_recv(newconn)) != NULL);

        if(vflag)
          printf("tcp-xm data closing...\n");

        /* close TCP-XM data, storing bytecounts before pcb deletion */
        netconn_close(newconn);
#ifdef TCPXM
        ubytesrcvd_early = newconn->pcb.tcp->ubytesrcvd;
#endif
        netconn_waitclose(newconn, 0, 2000);

#ifdef TCPXM
        ubytesrcvd = newconn->pcb.tcp->ubytesrcvd;
        mbytesrcvd = newconn->pcb.tcp->mbytesrcvd;
#endif

        /* use earlier bytecount in case pcb gone after waitclose */
        if(ubytesrcvd == 0) {
          ubytesrcvd = ubytesrcvd_early;
        }

        netconn_delete(newconn);
        netconn_delete(conn);

      } else {

        /* socket for incoming TCP data connection */ 
        if((s2 = socket(PF_INET, SOCK_STREAM, 0)) < 0){
          perror("can't create data socket");
          exit(1);
        }

        /* bind on any interface */
        bzero((char *) &serv_addr, sizeof(serv_addr));
        serv_addr.sin_family      = PF_INET;
        serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
        serv_addr.sin_port        = htons(mcp_dstport);

        if(bind(s2, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0){
          perror("can't bind local address");
          exit(1);
        }

        if(listen(s2, 1) <0) {
          perror("listen failed");
          exit(1);
        }

        if(vflag)
          printf("tcp data listening\n");

        if(!tflag) {
          /* tell client that we are ready to go */
          if(vflag)
            printf("sending ready to go\n");
          send_control(newsock, MCP_STATUS, MCP_OK);
        }

        clilen = sizeof(cli_addr);
        if((newsock2 = accept(s2, (struct sockaddr *) &cli_addr, &clilen))<0){
          perror("server: data accept error");
          exit(1);
        }

        if(vflag)
          printf("tcp data accepted\n");

        /* read data while connection open */
        bzero(data, sizeof(data));
        while((len = read(newsock2, data, sizeof(data))) > 0) {
          bytesrecd += len;
          if((r = write(fd, data, len)) < len) {
            sprintf(msg, "wrote just %d of %d bytes to file", r, len);
	    perror(msg);
          }
        }

        if(vflag)
          printf("tcp data closing...\n");

        /* close TCP data */
        close(newsock2);
        close(s2);

      } /* protocol */

      if(vflag)
        printf("data connection closed\n");

      close(fd);

      if(!tflag) {
        /* tell client how many bytes we received */
        send_control(newsock, MCP_FILESIZE, bytesrecd);
        send_control(newsock, MCP_FILESIZE, ubytesrcvd);
        send_control(newsock, MCP_FILESIZE, mbytesrcvd);

        /* close TCP control */
        if(vflag)
          printf("tcp control closing...\n");
        close(newsock);

        if(vflag)
          printf("tcp control closed\n");
      }

      if(vflag) {
        printf("%lu bytes received (lwIP layer: u=%u, m=%u)\n",
          bytesrecd, ubytesrcvd, mbytesrcvd);
        if(bytesrecd != mcpcntl.bytecount && !tflag) {
          printf("warning: %lu filesize != %lu bytes received\n", 
            mcpcntl.bytecount, bytesrecd);
        }
        printf("\n");
      }

      exit(0);

    } else { /* parent */

      /* avoid zombies */
#if defined(sun)
      if((void *)signal(SIGCHLD, SIG_IGN) == SIG_ERR) {
#else
      if((void *)signal(SIGCHLD, handler) == SIG_ERR) {
#endif
        perror("error initializing signal handler");
        exit(1);
      }

    } /* fork() */

    if(!tflag) {
      close(newsock);
    }

  } while(!tflag);

}

/* signal handler - allow child to die gracefully - avoiding zombies */
void handler()
{
  int status;
  wait(&status);  /* Wait for SIGCHLD */
}

usage() {
  printf("usage: mcpd [-hkKvz] [-d dir] [-p port] [-t <tapdev>]\n");
  exit(1);
}

helpusage() {
  printf("usage: mcpd [-hkKvz] [-d dir] [-p port] [-t <tapdev>]\n");
  printf("-d <dir>  == write files to specified directory\n");
  printf("-h        == this help\n");
  printf("-K        == kill all mcpd processes and exit\n");
  printf("-k        == -K, but run & kill children to get udp 2222 for xm\n");
  printf("-p <port> == tcp control port\n");
  printf("-t <dev#> == tap instead of udp\n");
  printf("-v        == verbose\n");
  printf("-z        == zap incoming data to /dev/null\n");
  exit(0);
}

/*
 * kill all other mcpd processes
*/
killmcpds(int vflag) {
  FILE *fp;
  char line[16];
  int pid, mypid, err;

  /* close stderr */
  err = dup(STDERR_FILENO);
  close(STDERR_FILENO);

  mypid = getpid();

  if(vflag)
    printf("mypid=%d, killing other mcpd processes\n", mypid);

#if defined(sun)
  fp = popen("ps -ef | grep -v 'mcpd -K' | awk '/mcpd/{print $2}'", "r");
#else
  fp = popen("ps ax | grep -v 'mcpd -K' | awk '/mcpd/{print $1}'", "r");
#endif

  while(fgets(line, sizeof(line), fp)) {
    pid = atoi(line);
    
    if(pid != mypid && pid > 0) {
      if(vflag)
        printf("killing %d\n", pid);
      kill(pid, SIGKILL);
    }
  }

  /* also kill mld */
#if defined(sun)
  system("pkill -9 mld > /dev/null 2>&1");
#else
  system("killall -9 mld > /dev/null 2>&1");
#endif

  fclose(fp);

  /* reopen stderr */
  dup2(err, STDERR_FILENO);
  close(err);
}
