#define __LWIP_INET_H__
#define UDPIF_XM

#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "lwip/debug.h"

#include "lwip/opt.h"
#include "lwip/def.h"
#include "lwip/ip.h"
#include "lwip/mem.h"
#include "lwip/pbuf.h"
#include "lwip/sys.h"

#include "lwip/netif.h"

#ifdef TCPXM
#include "lwip/igmp.h"
#endif

#define IFNAME0 'u'
#define IFNAME1 'd'



struct udpif {
  int fd;
  struct sockaddr_in saddr;
  struct sockaddr_in daddr;
};

/* Forward declarations. */
static void  udpif_input(struct netif *netif);
static err_t udpif_output(struct netif *netif, struct pbuf *p,
			       struct ip_addr *ipaddr);

static void udpif_thread(void *data);

/*-----------------------------------------------------------------------------------*/
static void
low_level_init(struct netif *netif)
{
  struct udpif *udpif;
  int sport, dport;
  int flag_on = 1;
  unsigned char mc_ttl=255;

  udpif = netif->state;

  /* extract source & dest ports */
  memcpy(&sport, &netif->netmask, sizeof(sport));
  memcpy(&dport, &netif->gw,      sizeof(dport));
  
  /* Create a UDP socket */
  udpif->fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);

  LWIP_DEBUGF(UDPIF_DEBUG, ("udpif_init: socket fd = %d\n", udpif->fd));
  if(udpif->fd == -1) {
    perror("udpif_init: error creating socket");
    exit(1);
  }

  /* set reuse port to on to allow multiple binds per host */
// kj: do i really want this?! commented out 14/4/04
/*
  if ((setsockopt(udpif->fd, SOL_SOCKET, SO_REUSEADDR, &flag_on,
       sizeof(flag_on))) < 0) {
    perror("udpif_init: setsockopt() failed on so_reuseaddr");
    exit(1);
  }
*/

  /* set the TTL (time to live/hop count) for the send */
  if ((setsockopt(udpif->fd, IPPROTO_IP, IP_MULTICAST_TTL,
       (void*) &mc_ttl, sizeof(mc_ttl))) < 0) {
    perror("udpif_init: setsockopt() failed on multicast ttl");
    exit(1);
  }

  // bind source address
  LWIP_DEBUGF(UDPIF_DEBUG, ("udpif_init: binding to socket\n"));
  bzero((char *)&udpif->saddr, sizeof(udpif->saddr));
  udpif->saddr.sin_family = PF_INET;
  udpif->saddr.sin_addr.s_addr = htonl(INADDR_ANY);
  udpif->saddr.sin_port = htons(sport);
  if(bind(udpif->fd,(struct sockaddr *)&udpif->saddr,sizeof(udpif->saddr))<0){
    perror("udpif_init: error binding");
    exit(1);
  }

  // dest address port/family
  udpif->daddr.sin_port = htons(dport);
  udpif->daddr.sin_family = PF_INET;

  sys_thread_new(udpif_thread, netif, DEFAULT_THREAD_PRIO);

}
/*-----------------------------------------------------------------------------------*/
/*
 * low_level_output():
 *
 * Should do the actual transmission of the packet. The packet is
 * contained in the pbuf that is passed to the function. This pbuf
 * might be chained.
 *
 */
/*-----------------------------------------------------------------------------------*/

static err_t
low_level_output(struct netif *netif, struct pbuf *p)
{
  struct pbuf *q;
  char buf[1500];
  char *bufptr;
  struct udpif *udpif;
  int bytes;

  udpif = netif->state;
  
  /* initiate transfer(); */
  
  bufptr = &buf[0];
  
  for(q = p; q != NULL; q = q->next) {
    /* Send the data from the pbuf to the interface, one pbuf at a
       time. The size of the data in each pbuf is kept in the ->len
       variable. */    
    /* send data from(q->payload, q->len); */
    memcpy(bufptr, q->payload, q->len);
    bufptr += q->len;
  }

  /* write to socket */
  bytes = sendto(udpif->fd, buf, p->tot_len, 0, 
      (struct sockaddr *)&udpif->daddr, sizeof(udpif->daddr));
  if((bytes != p->tot_len) || (p->tot_len > 1500)) {
    perror("udpif_low_level_output: sendto");
    perror(strerror(errno));
    printf("sock=%d, sendto of p->tot_len=%d bytes wrote %d bytes to %x\n",
      udpif->fd, p->tot_len, bytes, udpif->daddr.sin_addr);
  }
  LWIP_DEBUGF(UDPIF_DEBUG, ("udpif_low_level_output: wrote %d bytes\n",p->tot_len));

  return ERR_OK;
}
/*-----------------------------------------------------------------------------------*/
/*
 * low_level_input():
 *
 * Should allocate a pbuf and transfer the bytes of the incoming
 * packet from the interface into the pbuf.
 *
 */
/*-----------------------------------------------------------------------------------*/
static struct pbuf *
low_level_input(struct udpif *udpif)
{
  struct pbuf *p, *q;
  int len;
  char buf[1500];
  char *bufptr;

  /* Obtain packet size and put into the "len" variable. */
  len = recv(udpif->fd, buf, sizeof(buf), 0);

  LWIP_DEBUGF(UDPIF_DEBUG, ("udpif_low_level_input: read %d bytes\n", len));
  if(len <= 0) {
    return NULL;
  }
  
  /* We allocate a pbuf chain of pbufs from the pool. */
  //p = pbuf_alloc(PBUF_LINK, len, PBUF_POOL);
  p = pbuf_alloc(PBUF_RAW, len, PBUF_POOL);
  
  if(p != NULL) {
    /* We iterate over the pbuf chain until we have read the entire
       packet into the pbuf. */
    bufptr = &buf[0];
    for(q = p; q != NULL; q = q->next) {
      /* Read enough bytes to fill this pbuf in the chain. The
         avaliable data in the pbuf is given by the q->len
         variable. */
      /* read data into(q->payload, q->len); */
      memcpy(q->payload, bufptr, q->len);
      bufptr += q->len;
    }
    /* acknowledge that packet has been read(); */
  } else {
    /* drop packet(); */
    LWIP_DEBUGF(MEMP_DEBUG|2, ("udpif_low_level_input: pbuf_alloc was NULL\n"));
  }

  return p;  
}
/*-----------------------------------------------------------------------------------*/
static void 
udpif_thread(void *arg)
{
  struct netif *netif;
  struct udpif *udpif;
  fd_set fdset;
  int ret;
  
  netif = arg;
  udpif = netif->state;
  
  while(1) {
    FD_ZERO(&fdset);
    FD_SET(udpif->fd, &fdset);

    /* Wait for a packet to arrive. */
    ret = select(udpif->fd + 1, &fdset, NULL, NULL, NULL);

    if(ret == 1) {
      /* Handle incoming packet. */
      udpif_input(netif);
    } else if(ret == -1) {
      perror("udpif_thread: select");
    }
  }
}
/*-----------------------------------------------------------------------------------*/
/*
 * udpif_output():
 *
 * This function is called by the TCP/IP stack when an IP packet
 * should be sent. It calls the function called low_level_output() to
 * do the actuall transmission of the packet.
 *
 */
/*-----------------------------------------------------------------------------------*/
static err_t
udpif_output(struct netif *netif, struct pbuf *p,
		  struct ip_addr *ipaddr)
{
  struct udpif *udpif;
  struct ip_hdr *iphdr;
  struct ip_mreq imr;

  udpif = netif->state;

  iphdr = (struct ip_hdr *)(p->payload);

#ifdef TCPXM
  // first check if IGMP
  if(IPH_PROTO(iphdr) == IP_PROTO_IGMP) {

    // join group
    imr.imr_multiaddr.s_addr = iphdr->dest.addr;
    imr.imr_interface.s_addr = htonl(INADDR_ANY);

    if(setsockopt(udpif->fd, IPPROTO_IP, IP_ADD_MEMBERSHIP,
                       &imr, sizeof(struct ip_mreq)) < 0) {
      /*
      perror("udpif: can't join IGMP group!");
      exit(1);
      */
    }
    return ERR_OK;
  }
#endif

  // fill in destination address/port/family
  memcpy(&udpif->daddr.sin_addr, &ipaddr->addr, sizeof(ipaddr->addr));

  return low_level_output(netif, p);
}
/*-----------------------------------------------------------------------------------*/
/*
 * udpif_input():
 *
 * This function should be called when a packet is ready to be read
 * from the interface. It uses the function low_level_input() that
 * should handle the actual reception of bytes from the network
 * interface.
 *
 */
/*-----------------------------------------------------------------------------------*/
static void
udpif_input(struct netif *netif)
{
  struct udpif *udpif;
  struct pbuf *p;

  udpif = netif->state;
  
  p = low_level_input(udpif);

  if(p == NULL) {
    LWIP_DEBUGF(UDPIF_DEBUG, ("udpif_input: low_level_input returned NULL\n"));
    return;
  }

  // anything arriving on this port should be from another lwIP stack
  LWIP_DEBUGF(UDPIF_DEBUG, ("udpif_input: got packet\n"));
  netif->input(p, netif);
}
/*-----------------------------------------------------------------------------------*/
/*
 * udpif_init():
 *
 * Should be called at the beginning of the program to set up the
 * network interface. It calls the function low_level_init() to do the
 * actual setup of the hardware.
 *
 */
/*-----------------------------------------------------------------------------------*/
err_t
udpif_init(struct netif *netif)
{
  struct udpif *udpif;
    
  udpif = mem_malloc(sizeof(struct udpif));
  netif->state = udpif;
  netif->name[0] = IFNAME0;
  netif->name[1] = IFNAME1;
  netif->output = udpif_output;
  netif->linkoutput = low_level_output;
  
  low_level_init(netif);

  return ERR_OK;
}
/*-----------------------------------------------------------------------------------*/
